﻿using System;
using System.Collections.Generic;
using System.Diagnostics.Contracts;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using Derbster.DataAccess;
using Derbster.Domain;

namespace Derbster.Actions
{
    public class JamActions : IJamActions
    {
        private Jam _jam;
        private IDerbsterContext _derbsterContext;
        private IGameActionStack _actionStack;

        public JamActions(Jam jam, IDerbsterContext derbsterContext, IGameActionStack actionStack)
        {
            Contract.Requires(jam != null);
            Contract.Requires(derbsterContext != null);
            Contract.Requires(actionStack != null);

            _jam = jam;
            _derbsterContext = derbsterContext;
            _actionStack = actionStack;
        }

        public void FirstTeamScore(int score)
        {
            _jam.FirstTeamScore(score);
            this.TeamScoreAction(score, WhichTeam.First);
            _derbsterContext.SaveChanges();
        }

        public void SecondTeamScore(int score)
        {
            _jam.SecondTeamScore(score);
            this.TeamScoreAction(score, WhichTeam.Second);
            _derbsterContext.SaveChanges();
        }

        private void TeamScoreAction(int score, WhichTeam whichTeam)
        {
            Contract.Requires(_actionStack != null);

            GameAction<Jam> gameAction;
            if (whichTeam == WhichTeam.First)
            {
                gameAction = new GameAction<Jam>
                    {
                        DoAction = a => a.State.FirstTeamScore(score),
                        UndoAction = a => a.State.RemoveFirstTeamScoringPass(),
                        State = _jam
                    };
            }
            else
            {
                gameAction = new GameAction<Jam>
                    {
                        DoAction = a => a.State.SecondTeamScore(score),
                        UndoAction = a => a.State.RemoveSecondTeamScoringPass(),
                        State = _jam
                    };
            }
            _actionStack.Push(gameAction);
        }

        public void CorrectFirstTeamScore(int pass, int score)
        {
            int oldScore = _jam.CorrectFirstTeamScore(pass, score);
            this.CorrectTeamScoreAction(pass, score, oldScore, WhichTeam.First);
            _derbsterContext.SaveChanges();
        }

        public void CorrectSecondTeamScore(int pass, int score)
        {
            int oldScore = _jam.CorrectSecondTeamScore(pass, score);
            this.CorrectTeamScoreAction(pass, score, oldScore, WhichTeam.Second);
            _derbsterContext.SaveChanges();
        }

        private void CorrectTeamScoreAction(int pass, int score, int oldScore, WhichTeam whichTeam)
        {
            Contract.Requires(_actionStack != null);

            GameAction<Jam> gameAction;
            if (whichTeam == WhichTeam.First)
            {
                gameAction = new GameAction<Jam>
                {
                    DoAction = a => a.State.CorrectFirstTeamScore(pass, score),
                    UndoAction = a => a.State.CorrectFirstTeamScore(pass, oldScore),
                    State = _jam
                };
            }
            else
            {
                gameAction = new GameAction<Jam>
                {
                    DoAction = a => a.State.CorrectSecondTeamScore(pass, score),
                    UndoAction = a => a.State.CorrectSecondTeamScore(pass, oldScore),
                    State = _jam
                };
            }
            _actionStack.Push(gameAction);
        }

        public void RemoveFirstTeamScoringPass()
        {
            int score = _jam.RemoveFirstTeamScoringPass();
            this.RemoveTeamScoringPass(score, WhichTeam.First);
            _derbsterContext.SaveChanges();
        }

        public void RemoveSecondTeamScoringPass()
        {
            int score = _jam.RemoveSecondTeamScoringPass();
            this.RemoveTeamScoringPass(score, WhichTeam.Second);
            _derbsterContext.SaveChanges();
        }

        private void RemoveTeamScoringPass(int score, WhichTeam whichTeam)
        {
            Contract.Requires(_actionStack != null);

            GameAction<Jam> gameAction;
            if (whichTeam == WhichTeam.First)
            {
                gameAction = new GameAction<Jam>
                {
                    DoAction = a => a.State.RemoveFirstTeamScoringPass(),
                    UndoAction = a => a.State.FirstTeamScore(score),
                    State = _jam
                };
            }
            else
            {
                gameAction = new GameAction<Jam>
                {
                    DoAction = a => a.State.RemoveSecondTeamScoringPass(),
                    UndoAction = a => a.State.SecondTeamScore(score),
                    State = _jam
                };
            }
            _actionStack.Push(gameAction);
        }

        public void AddJammerToFirstTeamLineup(Skater skater)
        {
            if (_jam.FirstTeamLineup.Jammer != null) throw new IntegrityException("Lineup has a jammer.");
            _jam.FirstTeamLineup.Jammer = skater;
            this.AddSkaterToLineupAction(skater, SkaterPosition.Jammer, WhichTeam.First);
            _derbsterContext.SaveChanges();
        }

        public void AddPivotToFirstTeamLineup(Skater skater)
        {
            if (_jam.FirstTeamLineup.Pivot != null) throw new IntegrityException("Lineup has a pivot.");
            _jam.FirstTeamLineup.Pivot = skater;
            this.AddSkaterToLineupAction(skater, SkaterPosition.Pivot, WhichTeam.First);
            _derbsterContext.SaveChanges();
        }

        public void AddBlockerToFirstTeamLineup(Skater skater)
        {
            _jam.FirstTeamLineup.Blockers.Add(skater);
            this.AddSkaterToLineupAction(skater, SkaterPosition.Blocker, WhichTeam.First);
            _derbsterContext.SaveChanges();
        }

        public void AddJammerToSecondTeamLineup(Skater skater)
        {
            if (_jam.SecondTeamLineup.Jammer != null) throw new IntegrityException("Lineup has a jammer.");
            _jam.SecondTeamLineup.Jammer = skater;
            this.AddSkaterToLineupAction(skater, SkaterPosition.Jammer, WhichTeam.Second);
            _derbsterContext.SaveChanges();
        }

        public void AddPivotToSecondTeamLineup(Skater skater)
        {
            if (_jam.SecondTeamLineup.Pivot != null) throw new IntegrityException("Lineup has a pivot.");
            _jam.SecondTeamLineup.Pivot = skater;
            this.AddSkaterToLineupAction(skater, SkaterPosition.Pivot, WhichTeam.Second);
            _derbsterContext.SaveChanges();
        }

        public void AddBlockerToSecondTeamLineup(Skater skater)
        {
            _jam.SecondTeamLineup.Blockers.Add(skater);
            this.AddSkaterToLineupAction(skater, SkaterPosition.Blocker, WhichTeam.Second);
            _derbsterContext.SaveChanges();
        }

        private void AddSkaterToLineupAction(Skater skater, SkaterPosition position, WhichTeam whichTeam)
        {
            Contract.Requires(_actionStack != null);
            GameAction<Jam> gameAction;
            if (whichTeam == WhichTeam.First)
            {
                switch (position)
                {
                    case SkaterPosition.Blocker:
                        gameAction = new GameAction<Jam>
                        {
                            DoAction = a => a.State.FirstTeamLineup.Blockers.Add(skater),
                            UndoAction = a => a.State.RemoveFromFirstTeamLineup(skater),
                            State = _jam
                        };
                        break;
                    case SkaterPosition.Jammer:
                        gameAction = new GameAction<Jam>
                        {
                            DoAction = a => a.State.FirstTeamLineup.Jammer = skater,
                            UndoAction = a => a.State.RemoveFromFirstTeamLineup(skater),
                            State = _jam
                        };
                        break;
                    case SkaterPosition.Pivot:
                        gameAction = new GameAction<Jam>
                        {
                            DoAction = a => a.State.FirstTeamLineup.Pivot = skater,
                            UndoAction = a => a.State.RemoveFromFirstTeamLineup(skater),
                            State = _jam
                        };
                        break;
                    default:
                        gameAction = null;
                        break;
                }
            }
            else
            {
                switch (position)
                {
                    case SkaterPosition.Blocker:
                        gameAction = new GameAction<Jam>
                        {
                            DoAction = a => a.State.SecondTeamLineup.Blockers.Add(skater),
                            UndoAction = a => a.State.RemoveFromSecondTeamLineup(skater),
                            State = _jam
                        };
                        break;
                    case SkaterPosition.Jammer:
                        gameAction = new GameAction<Jam>
                        {
                            DoAction = a => a.State.SecondTeamLineup.Jammer = skater,
                            UndoAction = a => a.State.RemoveFromSecondTeamLineup(skater),
                            State = _jam
                        };
                        break;
                    case SkaterPosition.Pivot:
                        gameAction = new GameAction<Jam>
                        {
                            DoAction = a => a.State.SecondTeamLineup.Pivot = skater,
                            UndoAction = a => a.State.RemoveFromSecondTeamLineup(skater),
                            State = _jam
                        };
                        break;
                    default:
                        gameAction = null;
                        break;
                }
            }
            _actionStack.Push(gameAction);
        }

        public void CorrectFirstTeamLineup(Skater oldSkater, Skater newSkater)
        {
            _jam.CorrectFirstTeamLineup(oldSkater, newSkater);
            this.CorrectTeamLineupAction(oldSkater, newSkater, WhichTeam.First);
            _derbsterContext.SaveChanges();
        }

        public void CorrectSecondTeamLineup(Skater oldSkater, Skater newSkater)
        {
            _jam.CorrectSecondTeamLineup(oldSkater, newSkater);
            this.CorrectTeamLineupAction(oldSkater, newSkater, WhichTeam.Second);
            _derbsterContext.SaveChanges();
        }

        private void CorrectTeamLineupAction(Skater oldSkater, Skater newSkater, WhichTeam whichTeam)
        {
            Contract.Requires(_actionStack != null);

            GameAction<Jam> gameAction;
            if (whichTeam == WhichTeam.First)
            {
                gameAction = new GameAction<Jam>
                {
                    DoAction = a => a.State.CorrectFirstTeamLineup(oldSkater, newSkater),
                    UndoAction = a => a.State.CorrectFirstTeamLineup(newSkater, oldSkater),
                    State = _jam
                };
            }
            else
            {
                gameAction = new GameAction<Jam>
                {
                    DoAction = a => a.State.CorrectSecondTeamLineup(oldSkater, newSkater),
                    UndoAction = a => a.State.CorrectSecondTeamLineup(newSkater, oldSkater),
                    State = _jam
                };
            }
            _actionStack.Push(gameAction);
        }

        public void RemoveFromFirstTeamLineup(Skater skater)
        {
            SkaterPosition position;
            if (_jam.FirstTeamLineup.Jammer == skater) position = SkaterPosition.Jammer;
            else if (_jam.FirstTeamLineup.Pivot == skater) position = SkaterPosition.Pivot;
            else position = SkaterPosition.Blocker;

            _jam.RemoveFromFirstTeamLineup(skater);
            this.RemoveSkaterFromLineupAction(skater, position, WhichTeam.First);
            _derbsterContext.SaveChanges();
        }

        public void RemoveFromSecondTeamLineup(Skater skater)
        {
            SkaterPosition position;
            if (_jam.SecondTeamLineup.Jammer == skater) position = SkaterPosition.Jammer;
            else if (_jam.SecondTeamLineup.Pivot == skater) position = SkaterPosition.Pivot;
            else position = SkaterPosition.Blocker;

            this.RemoveSkaterFromLineupAction(skater, position, WhichTeam.First);
            _jam.RemoveFromSecondTeamLineup(skater);
            _derbsterContext.SaveChanges();
        }

        private void RemoveSkaterFromLineupAction(Skater skater, SkaterPosition position, WhichTeam whichTeam)
        {
            Contract.Requires(_actionStack != null);

            GameAction<Jam> gameAction;
            if (whichTeam == WhichTeam.First)
            {
                switch (position)
                {
                    case SkaterPosition.Blocker:
                        gameAction = new GameAction<Jam>
                        {
                            DoAction = a => a.State.RemoveFromFirstTeamLineup(skater),
                            UndoAction = a => a.State.FirstTeamLineup.Blockers.Add(skater),
                            State = _jam
                        };
                        break;
                    case SkaterPosition.Jammer:
                        gameAction = new GameAction<Jam>
                        {
                            DoAction = a => a.State.RemoveFromFirstTeamLineup(skater),
                            UndoAction = a => a.State.FirstTeamLineup.Jammer = skater,
                            State = _jam
                        };
                        break;
                    case SkaterPosition.Pivot:
                        gameAction = new GameAction<Jam>
                        {
                            DoAction = a => a.State.RemoveFromFirstTeamLineup(skater),
                            UndoAction = a => a.State.FirstTeamLineup.Pivot = skater,
                            State = _jam
                        };
                        break;
                    default:
                        gameAction = null;
                        break;
                }
            }
            else
            {
                switch (position)
                {
                    case SkaterPosition.Blocker:
                        gameAction = new GameAction<Jam>
                        {
                            DoAction = a => a.State.RemoveFromSecondTeamLineup(skater),
                            UndoAction = a => a.State.SecondTeamLineup.Blockers.Add(skater),
                            State = _jam
                        };
                        break;
                    case SkaterPosition.Jammer:
                        gameAction = new GameAction<Jam>
                        {
                            DoAction = a => a.State.RemoveFromSecondTeamLineup(skater),
                            UndoAction = a => a.State.SecondTeamLineup.Jammer = skater,
                            State = _jam
                        };
                        break;
                    case SkaterPosition.Pivot:
                        gameAction = new GameAction<Jam>
                        {
                            DoAction = a => a.State.RemoveFromSecondTeamLineup(skater),
                            UndoAction = a => a.State.SecondTeamLineup.Pivot = skater,
                            State = _jam
                        };
                        break;
                    default:
                        gameAction = null;
                        break;
                }
            }
            _actionStack.Push(gameAction);
        }

        public void FirstTeamIsLead()
        {
            _jam.FirstTeamIsLead();
            this.TeamIsLeadAction(WhichTeam.First);
            _derbsterContext.SaveChanges();
        }
       
        public void SecondTeamIsLead()
        {
            _jam.SecondTeamIsLead();
            this.TeamIsLeadAction(WhichTeam.Second);
            _derbsterContext.SaveChanges();
        }

        private void TeamIsLeadAction(WhichTeam whichTeam)
        {
            Contract.Requires(_actionStack != null);

            GameAction<Jam> gameAction;
            if (whichTeam == WhichTeam.First)
            {
                gameAction = new GameAction<Jam>
                {
                    DoAction = a => a.State.FirstTeamIsLead(),
                    UndoAction = a => a.State.FirstTeamScoring.Lead = false,
                    State = _jam
                };
            }
            else
            {
                gameAction = new GameAction<Jam>
                {
                    DoAction = a => a.State.SecondTeamIsLead(),
                    UndoAction = a => a.State.SecondTeamScoring.Lead = false,
                    State = _jam
                };
            }
            _actionStack.Push(gameAction);
        }

        public void FirstTeamLostLead()
        {
            _jam.FirstTeamLostLead();
            this.TeamLostLeadAction(WhichTeam.First);
            _derbsterContext.SaveChanges();
        }

        public void SecondTeamLostLead()
        {
            _jam.SecondTeamLostLead();
            this.TeamLostLeadAction(WhichTeam.Second);
            _derbsterContext.SaveChanges();
        }

        private void TeamLostLeadAction(WhichTeam whichTeam)
        {
            Contract.Requires(_actionStack != null);

            GameAction<Jam> gameAction;
            if (whichTeam == WhichTeam.First)
            {
                gameAction = new GameAction<Jam>
                {
                    DoAction = a => a.State.FirstTeamLostLead(),
                    UndoAction = a => a.State.FirstTeamScoring.Lost = false,
                    State = _jam
                };
            }
            else
            {
                gameAction = new GameAction<Jam>
                {
                    DoAction = a => a.State.SecondTeamLostLead(),
                    UndoAction = a => a.State.SecondTeamScoring.Lost = false,
                    State = _jam
                };
            }
            _actionStack.Push(gameAction);
        }

        public void JamCalled()
        {
            _jam.JamCalled();
            
            if (_jam.FirstTeamScoring.Called) this.JamCalledAction(WhichTeam.First);
            else this.JamCalledAction(WhichTeam.Second);

            _derbsterContext.SaveChanges();
        }

        private void JamCalledAction(WhichTeam whichTeam)
        {
            Contract.Requires(_actionStack != null);

            GameAction<Jam> gameAction;
            if (whichTeam == WhichTeam.First)
            {
                gameAction = new GameAction<Jam>
                {
                    DoAction = a => a.State.JamCalled(),
                    UndoAction = a => a.State.FirstTeamScoring.Called = false,
                    State = _jam
                };
            }
            else
            {
                gameAction = new GameAction<Jam>
                {
                    DoAction = a => a.State.JamCalled(),
                    UndoAction = a => a.State.SecondTeamScoring.Called = false,
                    State = _jam
                };
            }
            _actionStack.Push(gameAction);
        }

        public void FirstTeamStarPassed()
        {
            _jam.FirstTeamStarPassed();
            this.TeamStarPassedAction(WhichTeam.First);
            _derbsterContext.SaveChanges();
        }

        public void SecondTeamStarPassed()
        {
            _jam.SecondTeamStarPassed();
            this.TeamStarPassedAction(WhichTeam.Second);
            _derbsterContext.SaveChanges();
        }

        private void TeamStarPassedAction(WhichTeam whichTeam)
        {
            Contract.Requires(_actionStack != null);

            GameAction<Jam> gameAction;
            if (whichTeam == WhichTeam.First)
            {
                gameAction = new GameAction<Jam>
                    {
                        DoAction = a => a.State.FirstTeamStarPassed(),
                        UndoAction = a => a.State.RevokeFirstTeamStarPass(),
                        State = _jam
                    };
            }
            else
            {
                gameAction = new GameAction<Jam>
                    {
                        DoAction = a => a.State.FirstTeamStarPassed(),
                        UndoAction = a => a.State.RevokeFirstTeamStarPass(),
                        State = _jam
                    };
            }
            _actionStack.Push(gameAction);
        }

        public void RevokeFirstTeamStarPass()
        {
            _jam.RevokeFirstTeamStarPass();
            this.RevokeTeamStarPassedAction(WhichTeam.First);
            _derbsterContext.SaveChanges();
        }

        public void RevokeSecondTeamStarPass()
        {
            _jam.RevokeSecondTeamStarPass();
            this.RevokeTeamStarPassedAction(WhichTeam.Second);
            _derbsterContext.SaveChanges();
        }

        private void RevokeTeamStarPassedAction(WhichTeam whichTeam)
        {
            Contract.Requires(_actionStack != null);

            GameAction<Jam> gameAction;
            if (whichTeam == WhichTeam.First)
            {
                gameAction = new GameAction<Jam>
                {
                    DoAction = a => a.State.RevokeFirstTeamStarPass(),
                    UndoAction = a => a.State.FirstTeamStarPassed(),
                    State = _jam
                };
            }
            else
            {
                gameAction = new GameAction<Jam>
                {
                    DoAction = a => a.State.RevokeFirstTeamStarPass(),
                    UndoAction = a => a.State.FirstTeamStarPassed(),
                    State = _jam
                };
            }
            _actionStack.Push(gameAction);
        }
    }
}
