﻿using System;
using System.Collections.Generic;
using System.Diagnostics.Contracts;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using Derbster.DataAccess;
using Derbster.Domain;

namespace Derbster.Actions
{
    public class MatchActions : IMatchActions
    {
        private readonly IDerbsterContext _dbContext;
        private readonly IGameActionStack _actionStack;
        private readonly Match _match;

        public MatchActions(IDerbsterContext dbContext, IGameActionStack actionStack, Match match)
        {
            Contract.Requires(dbContext != null);
            Contract.Requires(actionStack != null);
            Contract.Requires(match != null);

            _dbContext = dbContext;
            _actionStack = actionStack;
            _match = match;
        }

        public Jam CurrentJam
        {
            get
            {
                return _match.CurrentJam;
            }
        }

        public bool JamInProgress
        {
            get
            {
                return _match.JamInProgress;
            }
        }

        public bool HasStarted
        {
            get
            {
                return _match.HasStarted;
            }
        }

        public bool HasFinished
        {
            get
            {
                return _match.HasFinished;
            }
        }

        public void StartMatch()
        {
            _match.StartMatch();
            this.StartMatchAction();
            SaveChanges();
        }

        private void StartMatchAction()
        {
            Contract.Requires(_actionStack != null);

            var newAction = new GameAction<Match>
                                {
                                    DoAction = a => a.State.StartMatch(),
                                    UndoAction = a => a.State.UndoStartMatch(),
                                    State = _match
                                };
            _actionStack.Push(newAction);
        }

        public void EndMatch()
        {
            _match.EndMatch();
            this.EndMatchAction();
            SaveChanges();
        }

        private void EndMatchAction()
        {
            Contract.Requires(_actionStack != null);

            var newAction = new GameAction<Match>
            {
                DoAction = a => a.State.EndMatch(),
                UndoAction = a => a.State.UndoEndMatch(),
                State = _match
            };
            _actionStack.Push(newAction);
        }

        public void StartPeriod()
        {
            _match.StartPeriod();
            this.StartPeriodAction();
            SaveChanges();
        }

        private void StartPeriodAction()
        {
            Contract.Requires(_actionStack != null);

            var newAction = new GameAction<Match>
            {
                DoAction = a => a.State.StartPeriod(),
                UndoAction = a => a.State.UndoStartPeriod(),
                State = _match
            };
            _actionStack.Push(newAction);
        }

        public void EndPeriod()
        {
            _match.EndPeriod();
            this.EndPeriodAction();
            SaveChanges();
        }

        private void EndPeriodAction()
        {
            Contract.Requires(_actionStack != null);

            var newAction = new GameAction<Match>
            {
                DoAction = a => a.State.EndPeriod(),
                UndoAction = a => a.State.UndoEndPeriod(),
                State = _match
            };
            _actionStack.Push(newAction);
        }

        public Jam StartJam(Lineup firstLineup, Lineup secondLineup)
        {
            Jam jam = _match.StartJam(firstLineup, secondLineup);
            this.StartJamAction(firstLineup, secondLineup);
            SaveChanges();
            return jam;
        }

        private void StartJamAction(Lineup firstLineup, Lineup secondLineup)
        {
            Contract.Requires(_actionStack != null);

            var newAction = new GameAction<Match>
            {
                DoAction = a => a.State.StartJam(firstLineup, secondLineup),
                UndoAction = a => a.State.UndoStartJam(),
                State = _match
            };
            _actionStack.Push(newAction);
        }

        public void EndJam()
        {
            _match.EndJam();
            this.EndJamAction();
            SaveChanges();
        }

        private void EndJamAction()
        {
            Contract.Requires(_actionStack != null);

            var newAction = new GameAction<Match>
            {
                DoAction = a => a.State.EndJam(),
                UndoAction = a => a.State.UndoEndJam(),
                State = _match
            };
            _actionStack.Push(newAction);
        }

        public Penalty SkaterPenalty(Skater skater, string code)
        {
            var penalty = _match.SkaterPenalty(skater, code);
            this.SkaterPenaltyAction(skater, code, penalty);
            SaveChanges();
            return penalty;
        }

        private void SkaterPenaltyAction(Skater skater, string code, Penalty penalty)
        {
            Contract.Requires(_actionStack != null);

            var newAction = new GameAction<Penalty>
            {
                DoAction = a => a.State = _match.SkaterPenalty(skater, code),
                UndoAction = a => _match.RemovePenalty(a.State),
                State = penalty
            };
            _actionStack.Push(newAction);
        }

        public void CorrectPenalty(Skater skater, int period, int jamNumber, string code)
        {
            _match.CorrectPenalty(skater, period, jamNumber, code);
            this.CorrectPenaltyAction(skater, period, jamNumber, code);
            SaveChanges();
        }

        private void CorrectPenaltyAction(Skater skater, int period, int jamNumber, string code)
        {
            Contract.Requires(_actionStack != null);

            var newAction = new GameAction<string>
            {
                DoAction = a => _match.CorrectPenalty(skater,period,jamNumber,a.State),
                UndoAction = a => _match.CorrectPenalty(skater,period,jamNumber,a.State),
                State = code
            };
            _actionStack.Push(newAction);
        }

        public void CorrectLastPenalty(Skater skater, string code)
        {
            _match.CorrectLastPenalty(skater, code);
            this.CorrectLastPenaltyAction(skater, code);
            SaveChanges();
        }

        private void CorrectLastPenaltyAction(Skater skater, string code)
        {
            Contract.Requires(_actionStack != null);

            var newAction = new GameAction<string>
            {
                DoAction = a => a.State = _match.CorrectLastPenalty(skater, a.State),
                UndoAction = a => a.State =_match.CorrectLastPenalty(skater, a.State),
                State = code
            };
            _actionStack.Push(newAction);
        }

        public void RemoveLastPenalty(Skater skater)
        {
            var oldCode = _match.RemoveLastPenalty(skater);
            this.RemoveLastPenaltyAction(skater, oldCode);
            SaveChanges();
        }

        private void RemoveLastPenaltyAction(Skater skater, string code)
        {
            Contract.Requires(_actionStack != null);

            var newAction = new GameAction<Match>
            {
                DoAction = a => a.State.SkaterPenalty(skater, code),
                UndoAction = a => a.State.RemoveLastPenalty(skater),
                State = _match
            };
            _actionStack.Push(newAction);
        }

        public void RemovePenalty(Penalty penalty)
        {
            _match.RemovePenalty(penalty);
            this.RemovePenaltyAction(penalty);
            SaveChanges();
        }

        private void RemovePenaltyAction(Penalty penalty)
        {
            Contract.Requires(_actionStack != null);

            var newAction = new GameAction<Match>
            {
                DoAction = a => a.State.RemovePenalty(penalty),
                UndoAction = a => a.State.AddPenalty(penalty),
                State = _match
            };
            _actionStack.Push(newAction);
        }

        public BoxEvent SkaterSeated(Skater skater)
        {
            var boxEvent = _match.SkaterSeated(skater);
            this.SkaterSeatedAction(skater, boxEvent);
            SaveChanges();
            return boxEvent;
        }

        private void SkaterSeatedAction(Skater skater, BoxEvent boxEvent)
        {
            Contract.Requires(_actionStack != null);

            var newAction = new GameAction<BoxEvent>
            {
                DoAction = a => a.State = _match.SkaterSeated(skater),
                UndoAction = a => _match.UndoBoxEvent(a.State),
                State = boxEvent
            };
            _actionStack.Push(newAction);
        }

        public BoxEvent SkaterReleased(Skater skater)
        {
            var boxEvent = _match.SkaterReleased(skater);
            this.SkaterReleasedAction(skater, boxEvent);
            SaveChanges();
            return boxEvent;
        }

        private void SkaterReleasedAction(Skater skater, BoxEvent boxEvent)
        {
            Contract.Requires(_actionStack != null);

            var newAction = new GameAction<BoxEvent>
            {
                DoAction = a => a.State = _match.SkaterReleased(skater),
                UndoAction = a => _match.UndoBoxEvent(a.State),
                State = boxEvent
            };
            _actionStack.Push(newAction);
        }

        public bool VerifyMatch()
        {
            return _match.VerifyMatch();
        }

        private int SaveChanges()
        {
            int count = _dbContext.SaveChanges();
            return count;
        }
    }
}
