﻿using System;
using System.Collections.Generic;
using System.Diagnostics.Contracts;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Derbster.DataAccess;
using Derbster.Domain;

namespace Derbster.Actions
{
    public class MatchManagementActions : IMatchManagementActions
    {
        private IDerbsterContext _dbContext;
        private IGameActionStack _actionStack;

        public MatchManagementActions(IDerbsterContext dbContext, IGameActionStack actionStack)
        {
            Contract.Requires(dbContext != null);
            Contract.Requires(actionStack != null);
            _dbContext = dbContext;
            _actionStack = actionStack;
        }

        public Match CreateMatch(Roster firstRoster, Roster secondRoster)
        {
            var newMatch = new Match(false, false, firstRoster, secondRoster);
            _dbContext.KnownRosters.Add(firstRoster);
            _dbContext.KnownRosters.Add(secondRoster);

            var count = _dbContext.KnownMatches.Count();
            newMatch.MatchId = String.Format("M-{0:D7}", count);
            _dbContext.KnownMatches.Add(newMatch);
            SaveChanges();

            this.CreateMatchAction(newMatch);
            return newMatch;
        }

        public Match CreateMatch(string name, Roster firstRoster, Roster secondRoster)
        {
            var newMatch = new Match(false, false, firstRoster, secondRoster, name);
        
            _dbContext.KnownRosters.Add(firstRoster);
            _dbContext.KnownRosters.Add(secondRoster);
            SaveChanges();

            var count = _dbContext.KnownMatches.Count();
            newMatch.MatchId = String.Format("M-{0:D7}", count);
            _dbContext.KnownMatches.Add(newMatch);
            SaveChanges();

            this.CreateMatchAction(newMatch);
            return newMatch;
        }

        public Match CreateRegulationMatch(Roster firstRoster, Roster secondRoster)
        {
            var newMatch = new Match(true, false, firstRoster, secondRoster);

            var count = _dbContext.KnownMatches.Count();
            newMatch.MatchId = String.Format("M-{0:D7}", count);
            _dbContext.KnownMatches.Add(newMatch);
            SaveChanges();

            this.CreateMatchAction(newMatch);
            return newMatch;
        }

        public Match CreateRegulationMatch(string name, Roster firstRoster, Roster secondRoster)
        {
            var newMatch = new Match(true, false, firstRoster, secondRoster, name);

            var count = _dbContext.KnownMatches.Count();
            newMatch.MatchId = String.Format("M-{0:D7}", count);
            _dbContext.KnownMatches.Add(newMatch);
            SaveChanges();

            this.CreateMatchAction(newMatch);
            return newMatch;
        }

        public Match CreateSanctionedMatch(Roster firstRoster, Roster secondRoster)
        {
            var newMatch = new Match(true, true, firstRoster, secondRoster);

            var count = _dbContext.KnownMatches.Count();
            newMatch.MatchId = String.Format("M-{0:D7}", count);
            _dbContext.KnownMatches.Add(newMatch);
            SaveChanges();

            this.CreateMatchAction(newMatch);
            return newMatch;
        }

        public Match CreateSanctionedMatch(string name, Roster firstRoster, Roster secondRoster)
        {
            var newMatch = new Match(true, true, firstRoster, secondRoster, name);

            var count = _dbContext.KnownMatches.Count();
            newMatch.MatchId = String.Format("M-{0:D7}", count);
            _dbContext.KnownMatches.Add(newMatch);
            SaveChanges();

            this.CreateMatchAction(newMatch);
            return newMatch;
        }

        internal Match RecreateMatch(Match match)
        {
            var existingMatch = _dbContext.KnownMatches.SingleOrDefault(m => m.MatchId == match.MatchId);
            if (existingMatch != null) return existingMatch;

            _dbContext.KnownMatches.Add(match);
            SaveChanges();

            return match;
        }

        private void CreateMatchAction(Match match)
        {
            Contract.Requires(match != null);
            _actionStack.Push(
                new GameAction<Match>
                    {
                        State = match,
                        DoAction = a => a.State = this.RecreateMatch(match),
                        UndoAction = a => this.DeleteMatch(match)
                    });
        }

        public Match FindMatchById(string matchId)
        {
            return _dbContext.KnownMatches.SingleOrDefault(m => m.MatchId == matchId);
        }

        public IEnumerable<Match> FindMatches(string keyword, bool matchWhole = false)
        {
            if (matchWhole)
                return _dbContext.
                    KnownMatches.Where(m => m.Name == keyword ||
                                              m.Where == keyword ||
                                              m.FirstTeamRoster.Name == keyword ||
                                              m.SecondTeamRoster.Name == keyword).AsEnumerable();

            return _dbContext.
                KnownMatches.Where(m => m.Name.Contains(keyword) ||
                                          m.Where.Contains(keyword) ||
                                          m.FirstTeamRoster.Name.Contains(keyword) ||
                                          m.SecondTeamRoster.Name.Contains(keyword)).AsEnumerable();
        }

        public IEnumerable<Match> FindMatchesByDate(DateTime fromDate, DateTime toDate)
        {
            return _dbContext.KnownMatches.Where(m => m.When >= fromDate && m.When <= toDate).AsEnumerable();
        }

        public IEnumerable<Match> FindMatchesByName(string name, bool matchWhole = false)
        {
            if (matchWhole)
                return _dbContext.KnownMatches.Where(m => m.Name == name).AsEnumerable();

            return _dbContext.KnownMatches.Where(m => m.Name.Contains(name)).AsEnumerable();
        }

        public IEnumerable<Match> FindMatchesByLocation(string location, bool matchWhole = false)
        {
            if (matchWhole)
                return _dbContext.KnownMatches.Where(m => m.Where == location).AsEnumerable();

            return _dbContext.KnownMatches.Where(m => m.Where.Contains(location)).AsEnumerable();
        }

        public void RenameMatch(Match match, string name)
        {
           
            if (_dbContext.KnownMatches.Any(m => m.MatchId == match.MatchId))
            {
                _dbContext.KnownMatches.Attach(match);
                var oldName = match.Name;
                match.Name = name;
                SaveChanges();

                this.RenameMatchAction(match, name, oldName);
            }
            else
            {
                throw new DerbsterActionException("Match was not found in database");
            }
        }

        public void RenameMatch(string matchId, string name)
        {
            var match = _dbContext.KnownMatches.SingleOrDefault(m => m.MatchId == matchId);
            if (match != null)
            {
                var oldName = match.Name;
                match.Name = name;
                SaveChanges();

                this.RenameMatchAction(match, name, oldName);
            }
            else
            {
                throw new DerbsterActionException("Match was not found in database");
            }
        }

        private void RenameMatchAction(Match match, string name, string oldName)
        {
            Contract.Requires(match != null);
            Contract.Requires(!string.IsNullOrWhiteSpace(name));
            Contract.Requires(!string.IsNullOrWhiteSpace(oldName));
            _actionStack.Push(
               new GameAction<string>
               {
                   State = match.MatchId,
                   DoAction = a => this.RenameMatch(a.State,name),
                   UndoAction = a => this.RenameMatch(a.State,oldName)
               });
        }

        [CannotUndo]
        public void DeleteMatch(Match match)
        {
            DeleteMatch(match.MatchId);
        }

        [CannotUndo]
        public void DeleteMatch(string matchId)
        {
            var match = _dbContext.KnownMatches.SingleOrDefault(m => m.MatchId == matchId);
            if (match == null)
            {
                throw new DerbsterActionException("Match was not found in database");
            }
            _dbContext.KnownMatches.Remove(match);
            SaveChanges();
        }

        private int SaveChanges()
        {
            int count = _dbContext.SaveChanges();
            return count;
        }
    }
}
