﻿using System;
using System.Collections.Generic;
using System.Diagnostics.Contracts;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Derbster.DataAccess;
using Derbster.Domain;

namespace Derbster.Actions
{
    public class PersonManagementActions : IPersonManagementActions
    {
        private IDerbsterContext _dbContext;
        private readonly IGameActionStack _actionStack;

        public PersonManagementActions(IDerbsterContext dbContext, IGameActionStack actionStack)
        {
            Contract.Requires(dbContext != null);
            Contract.Requires(actionStack != null);

            _dbContext = dbContext;
            _actionStack = actionStack;
        }

        public Skater CreateSkater(string name, string number)
        {
            Skater skater = this._dbContext.KnownSkaters.SingleOrDefault(s => s.Name == name && s.Number == number);
            if (skater == null)
            {
                var knownSkaterCount = _dbContext.KnownSkaters.Count();
                skater = new Skater(name, number)
                    {
                        SkaterId = String.Format("S-{0:D8}", knownSkaterCount)
                    };
                _dbContext.KnownSkaters.Add(skater);
                SaveChanges();
                CreateSkaterAction(skater, null);
            }

            return skater;
        }

        public Skater CreateSkater(string name, string number, Team team)
        {
            _dbContext.KnownTeams.Attach(team);
            Skater skater = this._dbContext.KnownSkaters.SingleOrDefault(s => s.Name == name && s.Number == number);
            if (skater == null)
            {
                var knownSkaterCount = _dbContext.KnownSkaters.Count();
                skater = new Skater(name, number)
                    {
                        SkaterId = String.Format("S-{0:D8}", knownSkaterCount)
                    };
            }
            if (!team.Skaters.Contains(skater))
            {
                _dbContext.KnownSkaters.Add(skater);
                team.Skaters.Add(skater);
                SaveChanges();
                CreateSkaterAction(skater, team);
            }
            return skater;
        }

        private void CreateSkaterAction(Skater skater, Team team)
        {
            Contract.Requires(_actionStack != null);

            if (team != null)
            {
                var newAction = new GameAction<Skater>
                                    {
                                        DoAction = a => a.State = this.CreateSkater(a.State.Name, a.State.Number, team),
                                        UndoAction = a => this.DeleteSkater(a.State),
                                        State = skater
                                    };
                _actionStack.Push(newAction);
            }
            else
            {
                var newAction = new GameAction<Skater>
                {
                    DoAction = a => a.State = this.CreateSkater(a.State.Name, a.State.Number),
                    UndoAction = a => this.DeleteSkater(a.State),
                    State = skater
                };
                _actionStack.Push(newAction);
            }
        }

        public Team CreateTeam(string name)
        {
            Team team = this._dbContext.KnownTeams.SingleOrDefault(t => t.Name == name);
            if (team == null)
            {
                var knownTeamCount = _dbContext.KnownTeams.Count();
                team = new Team(name)
                {
                    TeamId = String.Format("T-{0:D6}", knownTeamCount)
                };
                _dbContext.KnownTeams.Add(team);
                SaveChanges();
                CreateTeamAction(team);
            }

            return team;
        }

        public Team CreateTeam(string name, League league)
        {
            Team team = this._dbContext.KnownTeams.SingleOrDefault(t => t.Name == name);
            if (team == null)
            {
                var knownTeamCount = _dbContext.KnownTeams.Count();
                team = new Team(name)
                {
                    TeamId = String.Format("T-{0:D6}", knownTeamCount),
                    League = league
                };
                _dbContext.KnownTeams.Add(team);
                SaveChanges();
                CreateTeamAction(team);
            }
            return team;
        }

        private void CreateTeamAction(Team team)
        {
            Contract.Requires(_actionStack != null);
            var newAction = new GameAction<Team>
            {
                DoAction = a => a.State = this.CreateTeam(a.State.Name, a.State.League),
                UndoAction = a => this.DeleteTeam(a.State),
                State = team
            };
            _actionStack.Push(newAction);
        }

        public League CreateLeague(string name)
        {
            League league = this._dbContext.KnownLeagues.SingleOrDefault(l => l.Name == name);
            if (league == null)
            {
                var knownLeagueCount = _dbContext.KnownLeagues.Count();
                league = new League(name)
                    {
                        LeagueId = String.Format("L-{0:D6}", knownLeagueCount)
                    };
                _dbContext.KnownLeagues.Add(league);
                SaveChanges();
                CreateLeagueAction(league);
            }

            return league;
        }

        private void CreateLeagueAction(League league)
        {
            Contract.Requires(_actionStack != null);

            var newAction = new GameAction<League>
            {
                DoAction = a => a.State = this.CreateLeague(a.State.Name),
                UndoAction = a => this.DeleteLeague(a.State),
                State = league
            };
            _actionStack.Push(newAction);
        }

        public Official CreateOfficial(string name)
        {
            Official official = this._dbContext.KnownOfficials.
                SingleOrDefault(l => l.Name == name && l.Number == null);

            if (official == null)
            {
                var knownLeagueCount = _dbContext.KnownOfficials.Count();
                official = new Official(name)
                {
                    OfficialId = String.Format("O-{0:D8}", knownLeagueCount)
                };
                _dbContext.KnownOfficials.Add(official);
                SaveChanges();
                CreateOfficialAction(official);
            }
            return official;
        }

        public Official CreateOfficial(string name, string number)
        {
            Official official = this._dbContext.KnownOfficials.
                SingleOrDefault(l => l.Name == name && l.Number == number);
            if (official == null)
            {
                var knownLeagueCount = _dbContext.KnownOfficials.Count();
                official = new Official(name, number)
                {
                    OfficialId = String.Format("O-{0:D8}", knownLeagueCount)
                };
                _dbContext.KnownOfficials.Add(official);
                SaveChanges();
                CreateOfficialAction(official);
            }

            return official;
        }

        public Official CreateOfficial(string name, League league)
        {
            Official official = this._dbContext.KnownOfficials.
                SingleOrDefault(l => l.Name == name && l.Number == null);

            if (official == null)
            {
                var knownLeagueCount = _dbContext.KnownOfficials.Count();
                official = new Official(name)
                {
                    OfficialId = String.Format("O-{0:D8}", knownLeagueCount),
                    League = league
                };
                _dbContext.KnownOfficials.Add(official);
                SaveChanges();
                CreateOfficialAction(official);
            }

            return official;
        }

        public Official CreateOfficial(string name, string number, League league)
        {
            Official official = this._dbContext.KnownOfficials.
                SingleOrDefault(l => l.Name == name && l.Number == number);
            if (official == null)
            {
                var knownLeagueCount = _dbContext.KnownOfficials.Count();
                official = new Official(name, number)
                {
                    OfficialId = String.Format("O-{0:D8}", knownLeagueCount),
                    League = league
                };
                _dbContext.KnownOfficials.Add(official);
                SaveChanges();
                CreateOfficialAction(official);
            }

            return official;
        }

        private void CreateOfficialAction(Official official)
        {
            Contract.Requires(_actionStack != null);

            var newAction = new GameAction<Official>
            {
                DoAction = a => a.State = this.CreateOfficial(a.State.Name, a.State.Number, a.State.League),
                UndoAction = a => this.DeleteOfficial(a.State),
                State = official
            };
            _actionStack.Push(newAction);
        }

        public void ChangeSkaterName(Skater skater, string name)
        {
            if (!_dbContext.KnownSkaters.Any(s => s.SkaterId == skater.SkaterId))
                throw new ArgumentException("Unknown skater", "skater");

            _dbContext.KnownSkaters.Attach(skater);
            var oldName = skater.Name;
            skater.Name = name;
            SaveChanges();
            ChangeSkaterNameAction(skater, oldName, name);

        }

        private void ChangeSkaterNameAction(Skater skater, string oldName, string newName)
        {
            Contract.Requires(_actionStack != null);

            var newAction = new GameAction<Skater>
                {
                    State = skater,
                    DoAction = a => this.ChangeSkaterName(a.State, newName),
                    UndoAction = a => this.ChangeSkaterName(a.State, oldName)
                };
            _actionStack.Push(newAction);
        }

        public void ChangeSkaterNumber(Skater skater, string number)
        {
            if (!_dbContext.KnownSkaters.Any(s => s.SkaterId == skater.SkaterId))
                throw new ArgumentException("Unknown skater", "skater");

            _dbContext.KnownSkaters.Attach(skater);
            var oldNumber = skater.Number;
            skater.Number = number;
            SaveChanges();
            ChangeSkaterNumberAction(skater, oldNumber, number);
        }

        private void ChangeSkaterNumberAction(Skater skater, string oldNumber, string newNumber)
        {
            Contract.Requires(_actionStack != null);

            var newAction = new GameAction<Skater>
            {
                State = skater,
                DoAction = a => this.ChangeSkaterNumber(a.State, newNumber),
                UndoAction = a => this.ChangeSkaterNumber(a.State, oldNumber)
            };
            _actionStack.Push(newAction);
        }

        public void AddSkaterToTeam(Skater skater, Team team)
        {
            _dbContext.KnownSkaters.Attach(skater);
            _dbContext.KnownTeams.Attach(team);

            if (!_dbContext.KnownTeams.Any(t => t.Skaters.Contains(skater)))
                skater.Teams.Add(team);

            SaveChanges();
            AddSkaterTeamAction(skater, team);

        }

        public void RemoveSkaterFromTeam(Skater skater, Team team)
        {
            _dbContext.KnownSkaters.Attach(skater);
            _dbContext.KnownTeams.Attach(team);

            if (!_dbContext.KnownTeams.Any(t => t.Skaters.Contains(skater)))
                throw new IntegrityException("Skater not on team");

            team.Skaters.Remove(skater);
            SaveChanges();
            RemoveSkaterFromTeamAction(skater, team);

        }

        private void AddSkaterTeamAction(Skater skater, Team team)
        {
            Contract.Requires(_actionStack != null);

            var newAction = new GameAction<Skater>
            {
                State = skater,
                DoAction = a => this.AddSkaterToTeam(a.State, team),
                UndoAction = a => this.RemoveSkaterFromTeam(a.State, team)
            };
            _actionStack.Push(newAction);
        }

        private void RemoveSkaterFromTeamAction(Skater skater, Team team)
        {
            Contract.Requires(_actionStack != null);

            var newAction = new GameAction<Skater>
            {
                State = skater,
                DoAction = a => this.RemoveSkaterFromTeam(a.State, team),
                UndoAction = a => this.AddSkaterToTeam(a.State, team)
            };
            _actionStack.Push(newAction);
        }

        public void ChangeTeamName(Team team, string name)
        {
            if (!_dbContext.KnownTeams.Any(t => t.TeamId == team.TeamId))
                throw new ArgumentException("Unknown team", "team");

            _dbContext.KnownTeams.Attach(team);
            var oldName = team.Name;
            team.Name = name;
            SaveChanges();
            ChangeTeamNameAction(team, oldName, name);

        }

        private void ChangeTeamNameAction(Team team, string oldName, string newName)
        {
            Contract.Requires(_actionStack != null);

            var newAction = new GameAction<Team>
            {
                State = team,
                DoAction = a => this.ChangeTeamName(a.State, newName),
                UndoAction = a => this.ChangeTeamName(a.State, oldName)
            };
            _actionStack.Push(newAction);
        }

        public void ChangeLeagueName(League league, string name)
        {
            if (!_dbContext.KnownLeagues.Any(l => l.LeagueId == league.LeagueId))
                throw new ArgumentException("Unknown league", "league");

            _dbContext.KnownLeagues.Attach(league);
            var oldName = league.Name;
            league.Name = name;
            SaveChanges();
            ChangeLeagueNameAction(league, oldName, name);
        }

        private void ChangeLeagueNameAction(League league, string oldName, string newName)
        {
            Contract.Requires(_actionStack != null);

            var newAction = new GameAction<League>
            {
                State = league,
                DoAction = a => this.ChangeLeagueName(a.State, newName),
                UndoAction = a => this.ChangeLeagueName(a.State, oldName)
            };
            _actionStack.Push(newAction);
        }

        [CannotUndo]
        public void DeleteOfficial(Official official)
        {
            _dbContext.KnownOfficials.Attach(official);
            if (!_dbContext.KnownOfficials.Contains(official))
            {
                throw new IntegrityException("Unknown Official.");
            }
            _dbContext.KnownOfficials.Remove(official);
            SaveChanges();

        }

        [CannotUndo]
        public void DeleteSkater(Skater skater)
        {
            if (!_dbContext.KnownSkaters.Any(s => s.SkaterId == skater.SkaterId))
                throw new ArgumentException("Unknown skater", "skater");

            _dbContext.KnownSkaters.Attach(skater);
            _dbContext.KnownSkaters.Remove(skater);
            SaveChanges();

        }

        [CannotUndo]
        public void DeleteTeam(Team team)
        {
            if (!_dbContext.KnownTeams.Any(t => t.TeamId == team.TeamId))
                throw new ArgumentException("Unknown team", "team");

            _dbContext.KnownTeams.Attach(team);
            _dbContext.KnownTeams.Remove(team);
            SaveChanges();

        }

        public void DeleteLeague(League league)
        {
            if (!_dbContext.KnownLeagues.Any(l => l.LeagueId == league.LeagueId))
                throw new ArgumentException("Unknown league", "league");

            _dbContext.KnownLeagues.Attach(league);
            _dbContext.KnownLeagues.Remove(league);
            SaveChanges();
        }

        private int SaveChanges()
        {
            return _dbContext.SaveChanges();
        }
    }
}
