﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Diagnostics.Contracts;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using Derbster.Actions;
using Derbster.DataAccess;
using Derbster.Domain;

namespace Derbster.Console
{
    public class GameGenerator
    {
        private readonly IGameActionStack _actionStack;
        private readonly IDerbsterContext _dbContext;
        private readonly IPersonManagementActions _personManagementActions;
        private readonly IMatchManagementActions _matchManagementActions;

        private readonly double _penaltyProbability;

        private IMatchActions _matchActions;
        private IJamActions _jamActions;
        private Roster _firstTeamRoster;
        private Roster _secondTeamRoster;

        private const string Codes = "ABCEFGHILMNOPSXZ";

        private readonly string[] _firstNames =
            {
                "Slayer",
                "Picker",
                "Smacker",
                "Testher",
                "Destroyher",
                "Rollya",
                "Moher",
                "Flickher",
                "Effin",
                "Killya",
                "Plauger",
                "Plower",
                "Wrecker",
                "Juker",
                "Bringer",
            };


        private readonly string[] _lastNames =
            {
                "On Track",
                "All Out",
                "Will",
                "Deadly",
                "Nothingness",
                "Speed Demon",
                "'n' Leave Em",
                "Of Sorrows",
                "And Run",
                "Chaos",
                "Rainbows",
                "Managled",
                "Missunderstood",
                "Disorder",
                "Sadness"
            };

       

        public GameGenerator(IGameActionStack actionStack, IDerbsterContext dbContext, double penaltyProbability)
        {
            _actionStack = actionStack;
            _dbContext = dbContext;
            _personManagementActions = new PersonManagementActions(_dbContext, _actionStack);
            _matchManagementActions = new MatchManagementActions(_dbContext, _actionStack);
            _penaltyProbability = penaltyProbability;
        }

        public Match SimulateMatch()
        {
            var match = GenerateRandomMatch();
            _matchActions = new MatchActions(_dbContext, _actionStack, match);
            _matchActions.StartMatch();
            _matchActions.StartPeriod();
            this.SimulatePeriod();
            _matchActions.EndPeriod();
            _matchActions.StartPeriod();
            this.SimulatePeriod();
            _matchActions.EndPeriod();
            _matchActions.EndMatch();

            return match;
        }

        private void SimulatePeriod()
        {
            var random = new Random();

            Jam lastJam = null;

            for (int i = 0; i < random.Next(15, 23); i++)
            {
                var firstLineup = GenerateTeamRandomLineup(true, lastJam);
                var secondLineup = GenerateTeamRandomLineup(false, lastJam);
                lastJam = _matchActions.StartJam(firstLineup, secondLineup);
                SimulateJam();
                _matchActions.EndJam();
            }
        }

        private void SimulateJam()
        {
            var random = new Random();
            int firstTeamPasses = random.Next(0, 6);
            int secondTeamPasses = random.Next(0, 6);

            // Phase one. First pass.

            var firstHasPenalty = (random.NextDouble() <= _penaltyProbability);
            var secondHasPenalty = (random.NextDouble() <= _penaltyProbability);

            _jamActions = new JamActions(_matchActions.CurrentJam, _dbContext, _actionStack);

            if (firstHasPenalty)
            {
                AssignRandomPenalty(_matchActions.CurrentJam.FirstTeamLineup);
            }

            if (secondHasPenalty)
            {
                AssignRandomPenalty(_matchActions.CurrentJam.SecondTeamLineup);
            }

            for (int i = 0; i < random.Next(3,5); i++)
            {
                if (_matchActions.CurrentJam.FirstTeamScoring.Passes.Count < firstTeamPasses)
                {
                    if (_matchActions.CurrentJam.FirstTeamLineup.Jammer.State == SkaterState.Ready)
                        _jamActions.FirstTeamScore(random.Next(1, 5));
                }
                if (_matchActions.CurrentJam.SecondTeamScoring.Passes.Count < secondTeamPasses)
                {
                    if (_matchActions.CurrentJam.SecondTeamLineup.Jammer.State == SkaterState.Ready)
                        _jamActions.SecondTeamScore(random.Next(1, 5));
                }

                firstHasPenalty = (random.NextDouble() <= _penaltyProbability);
                secondHasPenalty = (random.NextDouble() <= _penaltyProbability);

                if (firstHasPenalty)
                {
                    AssignRandomPenalty(_matchActions.CurrentJam.FirstTeamLineup);
                }

                if (secondHasPenalty)
                {
                    AssignRandomPenalty(_matchActions.CurrentJam.SecondTeamLineup);
                }
            }

            foreach (var skater in _matchActions.CurrentJam.FirstTeamLineup)
            {
                if (skater.State == SkaterState.InBox)
                {
                    _matchActions.SkaterReleased(skater);
                }
            }

            foreach (var skater in _matchActions.CurrentJam.SecondTeamLineup)
            {
                if (skater.State == SkaterState.InBox)
                {
                    _matchActions.SkaterReleased(skater);
                }
            }
        }

        private void AssignRandomPenalty(Lineup currentLineup)
        {
            Contract.Requires(currentLineup != null);

            var random = new Random();
            var r = random.NextDouble();

            var randomCode = Codes[random.Next(0, Codes.Length)].ToString();
            if (r < .10)
            {
                if (currentLineup.Jammer.State == SkaterState.Ready)
                {
                    _matchActions.SkaterPenalty(currentLineup.Jammer, randomCode);
                    _matchActions.SkaterSeated(currentLineup.Jammer);
                }
            }
            else if (r > .10 && r < .30)
            {
                if (currentLineup.Pivot != null && currentLineup.Pivot.State == SkaterState.Ready)
                {
                    _matchActions.SkaterPenalty(currentLineup.Pivot, randomCode);
                    _matchActions.SkaterSeated(currentLineup.Pivot);
                }
                else if (currentLineup.Pivot == null)
                {
                    var blockers = currentLineup.Blockers.Where(b => b.State == SkaterState.Ready).ToArray();
                    if (blockers.Length > 1)
                    {
                        var whichBlocker = blockers.ElementAt(random.Next(0, blockers.Length));
                        _matchActions.SkaterPenalty(whichBlocker, randomCode);
                        _matchActions.SkaterSeated(whichBlocker);
                    }
                }
            }
            else
            {
                var blockers = currentLineup.Blockers.Where(b => b.State == SkaterState.Ready).ToArray();
                if (blockers.Length > 1)
                {
                    var whichBlocker = blockers.ElementAt(random.Next(0, blockers.Length));
                    _matchActions.SkaterPenalty(whichBlocker, randomCode);
                    _matchActions.SkaterSeated(whichBlocker);
                }
            }
        }

        private HashSet<Skater> CreateNewSkatersForLineup(Roster roster, Lineup lastLineup, int count)
        {
            Contract.Requires(roster != null);
            Contract.Requires(lastLineup != null); 
            HashSet<Skater> skaters = new HashSet<Skater>();
            foreach (var skater in lastLineup.Blockers.Where(s => s.State != SkaterState.Ready))
            {
                skaters.Add(skater);
            }
            if (lastLineup.Jammer.State != SkaterState.Ready)
                skaters.Add(lastLineup.Jammer);
            if (lastLineup.Pivot.State != SkaterState.Ready)
                skaters.Add(lastLineup.Pivot);

            var random = new Random();
            while (skaters.Count < count)
            {
                var newBlocker = roster.Members.ElementAt(random.Next(0, roster.Members.Count));
                if (newBlocker.State != SkaterState.Ready || skaters.Contains(newBlocker))
                    continue;

                skaters.Add(newBlocker);
            }
            return skaters;
        }

        private Lineup GenerateTeamRandomLineup(bool isFirstTeam, Jam lastJam)
        {
            Lineup newLineup;
            var roster = isFirstTeam ? _firstTeamRoster : _secondTeamRoster;

            if (lastJam != null)
            {
                var lastLineup = isFirstTeam ? lastJam.FirstTeamLineup : lastJam.SecondTeamLineup;

                if (lastLineup.Jammer.State != SkaterState.Ready)
                {
                    if (lastLineup.Pivot.State != SkaterState.Ready)
                    {
                        newLineup = new Lineup(lastLineup.Jammer, lastLineup.Pivot,
                                               CreateNewSkatersForLineup(roster, lastLineup, 3));
                    }
                    else
                    {
                        var skaters = CreateNewSkatersForLineup(roster, lastLineup, 4);
                        var pivot = skaters.First();
                        skaters.Remove(pivot);
                        newLineup = new Lineup(lastLineup.Jammer, pivot, skaters);    
                    }
                }
                else if (lastLineup.Pivot.State != SkaterState.Ready)
                {
                    var skaters = CreateNewSkatersForLineup(roster, lastLineup, 4);
                    var jammer = skaters.First();
                    skaters.Remove(jammer);
                    newLineup = new Lineup(jammer, lastLineup.Pivot, skaters);
                }
                else
                {
                    var skaters = CreateNewSkatersForLineup(roster, lastLineup, 5);
                    var jammer = skaters.First();
                    skaters.Remove(jammer);
                    var pivot = skaters.First();
                    skaters.Remove(pivot);
                    newLineup = new Lineup(jammer, pivot, skaters);
                }
            }
            else
            {
                var skaters = new HashSet<Skater>();
                while (skaters.Count < 5)
                {
                    var random = new Random();
                    var newSkater = roster.Members.ElementAt(random.Next(0, roster.Members.Count()));
                    if (skaters.Contains(newSkater) || newSkater.State != SkaterState.Ready)
                        continue;

                    skaters.Add(newSkater);
                }
                var jammer = skaters.First();
                skaters.Remove(jammer);
                var pivot = skaters.First();
                skaters.Remove(pivot);
                newLineup = new Lineup(jammer, pivot, skaters);
            }
            return newLineup;
        }

        private Match GenerateRandomMatch()
        {
           
            var random = new Random();

            Team firstTeam = _personManagementActions.CreateTeam(String.Format("Team {0:D6}", random.Next(0, 999999)));
            Team secondTeam = _personManagementActions.CreateTeam(String.Format("Team {0:D6}", random.Next(0, 999999)));

            _firstTeamRoster = new Roster("First","Black");
            _secondTeamRoster = new Roster("Second","White");
            
            while (_firstTeamRoster.Members.Count < 14)
            {
                var name = _firstNames[random.Next(0, _firstNames.Length)] + " " +
                           _lastNames[random.Next(0, _lastNames.Length)];
                var number = string.Format("{0:D4}", random.Next(0, 9999));
                if (_firstTeamRoster.Members.Any(s => s.Name == name || s.Number == number))
                    continue;

                _firstTeamRoster.Members.Add(_personManagementActions.CreateSkater(name, number, firstTeam));
            }

            Contract.Assume(_firstTeamRoster.Members.Count == 14);

            _firstTeamRoster.Captain = _firstTeamRoster.Members.ElementAt(0);
            _firstTeamRoster.Alternate = _firstTeamRoster.Members.ElementAt(1);

            while (_secondTeamRoster.Members.Count < 14)
            {
                var name = _firstNames[random.Next(0, _firstNames.Length)] + " " +
                           _lastNames[random.Next(0, _lastNames.Length)];
                var number = string.Format("{0:D4}", random.Next(0, 9999));
                if (_secondTeamRoster.Members.Any(s => s.Name == name || s.Number == number))
                    continue;

                _secondTeamRoster.Members.Add(_personManagementActions.CreateSkater(name, number, secondTeam));
            }

            Contract.Assume(_secondTeamRoster.Members.Count == 14);

            _secondTeamRoster.Captain = _secondTeamRoster.Members.ElementAt(0);
            _secondTeamRoster.Alternate = _secondTeamRoster.Members.ElementAt(1);

            return _matchManagementActions.CreateMatch(string.Format("Match {0:D5}", random.Next(0, 100000)), _firstTeamRoster,
                                                _secondTeamRoster);  
        }
    }
}
