﻿using System;
using System.Collections.Generic;
using System.ComponentModel.DataAnnotations;
using System.ComponentModel.DataAnnotations.Schema;
using System.Diagnostics.Contracts;
using System.Linq;
using System.Text;

namespace Derbster.Domain
{
    public class Jam
    {
        [ContractVerification(false)]
        internal Jam()
        {

        }

        public Jam(Match match, int period, int number, Lineup firstLineup, Lineup secondLineup)
        {
            Contract.Requires(match != null);
            Contract.Requires(period == 1 || period == 2);
            Contract.Requires(number > 0);
            Contract.Requires(firstLineup != null);
            Contract.Requires(secondLineup != null);
            Contract.Ensures(FirstTeamLineup != null);
            Contract.Ensures(SecondTeamLineup != null);

            Period = period;
            Number = number;
            Match = match;
            FirstTeamLineup = firstLineup;
            SecondTeamLineup = secondLineup;
            if (FirstTeamLineup.Jammer != null)
                FirstTeamScoring = new Scoring(FirstTeamLineup.Jammer);
            if (SecondTeamLineup.Jammer != null)
                SecondTeamScoring = new Scoring(SecondTeamLineup.Jammer);   
        }

        [Key, DatabaseGenerated(DatabaseGeneratedOption.Identity)]
        public int JamId { get; set; }

        public string MatchId { get; internal set; }
        public Match Match { get; internal set; }

        public int Period { get; internal set; }
        public int Number { get; internal set; }

        public Lineup FirstTeamLineup { get; internal set; }
        public Lineup SecondTeamLineup { get; internal set; }

        public Scoring FirstTeamScoring { get; internal set; }
        public Scoring SecondTeamScoring { get; internal set; }
        
        public void FirstTeamScore(int score)
        {
            Contract.Requires(score >= 1 && score <= 6);
            Contract.Requires(FirstTeamScoring != null);
            Contract.Requires(FirstTeamScoring.Passes != null);
            TeamScore(FirstTeamScoring, score);
        }

        public void SecondTeamScore(int score)
        {
            Contract.Requires(score >= 1 && score <= 6);
            Contract.Requires(SecondTeamScoring != null);
            Contract.Requires(SecondTeamScoring.Passes != null);
            TeamScore(SecondTeamScoring, score);
        }

        private void TeamScore(Scoring scoring, int score)
        {
            Contract.Requires(scoring != null);
            Contract.Requires(scoring.Passes != null);
            Contract.Requires(score >= 0 && score <= 6);
            int currentPass = scoring.Passes.Count;
            scoring.Passes.Add(new ScoringPass
            {
                Pass = currentPass + 1,
                Score = score
            });
        }

        public int CorrectFirstTeamScore(int pass, int score)
        {
            Contract.Requires(score >= 0 && score <= 6);
            Contract.Requires(FirstTeamScoring != null);
            return CorrectTeamScore(FirstTeamScoring, pass, score);
        }

        public int CorrectSecondTeamScore(int pass, int score)
        {
            Contract.Requires(score >= 0 && score <= 6);
            Contract.Requires(SecondTeamScoring != null);
            return CorrectTeamScore(SecondTeamScoring, pass, score);
        }

        private int CorrectTeamScore(Scoring scoring, int pass, int newScore)
        {
            Contract.Requires(scoring != null);
            Contract.Requires(scoring.Passes != null);
            Contract.Requires(newScore >= 0 && newScore <= 6);
            if (pass > scoring.Passes.Count)
                throw new ArgumentOutOfRangeException("pass", pass, "Invalid Scoring Pass Number");
           

            var correctedPass = scoring.Passes.SingleOrDefault(sp => sp.Pass == pass);
            if (correctedPass == null)
                throw new IntegrityException("Scoring");

            var oldScore = correctedPass.Score;
            correctedPass.Score = newScore;
            return oldScore;
        }

        public int RemoveFirstTeamScoringPass()
        {
            Contract.Requires(FirstTeamScoring != null);
            Contract.Requires(FirstTeamScoring.Passes != null);
            Contract.Requires(FirstTeamScoring.Passes.Count > 0);
            return RemoveTeamScoringPass(FirstTeamScoring);
        }

        public int RemoveSecondTeamScoringPass()
        {
            Contract.Requires(SecondTeamScoring != null);
            Contract.Requires(SecondTeamScoring.Passes != null);
            Contract.Requires(SecondTeamScoring.Passes.Count > 0);
            return RemoveTeamScoringPass(SecondTeamScoring);
        }
       
        private int RemoveTeamScoringPass(Scoring scoring)
        {
            Contract.Requires(scoring != null);
            Contract.Requires(scoring.Passes != null);
            Contract.Requires(scoring.Passes.Count > 0);

            var pass = scoring.Passes.Single(sp=>sp.Pass == scoring.Passes.Max(sp1 => sp1.Pass));
            var oldScore = pass.Score;
            scoring.Passes.Remove(pass);
            return oldScore;
        }

        public void CorrectFirstTeamLineup(Skater oldSkater, Skater newSkater)
        {
            Contract.Requires(oldSkater != null);
            Contract.Requires(newSkater != null);
            Contract.Requires(FirstTeamLineup != null);
            CorrectTeamLineup(FirstTeamLineup, oldSkater, newSkater);
        }

        public void CorrectSecondTeamLineup(Skater oldSkater, Skater newSkater)
        {
            Contract.Requires(oldSkater != null);
            Contract.Requires(newSkater != null);
            Contract.Requires(SecondTeamLineup != null);
            CorrectTeamLineup(SecondTeamLineup, oldSkater, newSkater);
        }

        private void CorrectTeamLineup(Lineup lineup, Skater oldSkater, Skater newSkater)
        {
            Contract.Requires(lineup != null);
            Contract.Requires(lineup.Blockers != null);
            Contract.Requires(oldSkater != null);
            Contract.Requires(newSkater != null);

            if (lineup.Jammer != null && lineup.Jammer == oldSkater)
            {
                lineup.Jammer = newSkater;
            }
            if (lineup.Pivot != null && lineup.Pivot == oldSkater)
            {
                lineup.Pivot = newSkater;
            }
            else if (lineup.Blockers == null || lineup.Blockers.Count == 0)
                throw new ArgumentOutOfRangeException("oldSkater", oldSkater, "Skater was not part of lineup");

            var wasInLineup = lineup.Blockers.Remove(oldSkater);

            if (!wasInLineup)
                throw new ArgumentOutOfRangeException("oldSkater", oldSkater, "Skater was not part of lineup");
        }

        public void RemoveFromFirstTeamLineup(Skater skater)
        {
            Contract.Requires(skater != null);
            Contract.Requires(FirstTeamLineup != null);
            RemoveFromTeamLineup(FirstTeamLineup, skater);
        }

        public void RemoveFromSecondTeamLineup(Skater skater)
        {
            Contract.Requires(skater != null);
            Contract.Requires(SecondTeamLineup != null);
            RemoveFromTeamLineup(SecondTeamLineup, skater);
        }

        private void RemoveFromTeamLineup(Lineup lineup, Skater skater)
        {
            Contract.Requires(lineup != null);
            Contract.Requires(lineup.Blockers != null);
            Contract.Requires(skater != null);
            
            if (lineup.Jammer != null && lineup.Jammer == skater)
            {
                lineup.Jammer = null;
            }
            if (lineup.Pivot != null && lineup.Pivot == skater)
            {
                lineup.Pivot = null;
            }
            else if (lineup.Blockers == null || lineup.Blockers.Count == 0)
                throw new ArgumentOutOfRangeException("skater", skater, "Skater was not part of lineup");

            var wasInLineup = lineup.Blockers.Remove(skater);
            if (!wasInLineup)
                throw new ArgumentOutOfRangeException("skater", skater, "Skater was not part of lineup");
        }

        public void FirstTeamIsLead()
        {
            Contract.Requires(FirstTeamScoring != null);
            FirstTeamScoring.Lead = true;
        }

        public void FirstTeamLostLead()
        {
            Contract.Requires(FirstTeamScoring != null);
            FirstTeamScoring.Lost = true;
        }

        public void SecondTeamIsLead()
        {
            Contract.Requires(FirstTeamScoring != null);
            FirstTeamScoring.Lead = true;
        }

        public void SecondTeamLostLead()
        {
            Contract.Requires(FirstTeamScoring != null);
            FirstTeamScoring.Lost = true;
        }

        public void JamCalled()
        {
            Contract.Requires(FirstTeamScoring != null);
            Contract.Requires(SecondTeamScoring != null);
            Contract.Requires(FirstTeamScoring.Lead || SecondTeamScoring.Lead);

            if (FirstTeamScoring.Lead) FirstTeamScoring.Called = true;
            if (SecondTeamScoring.Lead) SecondTeamScoring.Called = true;
        }

        public void FirstTeamStarPassed()
        {
            Contract.Requires(FirstTeamLineup != null);
            Contract.Requires(FirstTeamScoring != null);
            Contract.Requires(FirstTeamScoring.Passes != null);
            Contract.Requires(FirstTeamLineup.Pivot != null);
            Contract.Requires(FirstTeamLineup.Jammer != null);
            TeamStarPassed(FirstTeamLineup, FirstTeamScoring);
        }

        public void SecondTeamStarPassed()
        {
            Contract.Requires(SecondTeamLineup != null);
            Contract.Requires(SecondTeamScoring != null);
            Contract.Requires(SecondTeamScoring.Passes != null);
            Contract.Requires(SecondTeamLineup.Pivot != null);
            Contract.Requires(SecondTeamLineup.Jammer != null);
            TeamStarPassed(SecondTeamLineup, SecondTeamScoring);
        }

        private void TeamStarPassed(Lineup lineup, Scoring scoring)
        {
            Contract.Requires(lineup != null);
            Contract.Requires(scoring != null);
            Contract.Requires(scoring.Passes != null);
            Contract.Requires(lineup.Pivot != null);
            Contract.Requires(lineup.Jammer != null);

            lineup.StarPassed = true;
            scoring.StarPassedDuringPass = scoring.Passes.Count == 0 ? 0 : scoring.Passes.Max(sp => sp.Pass);
            Contract.Assert(scoring.StarPassedDuringPass.HasValue && scoring.StarPassedDuringPass >= 0);
            scoring.StarPassed = true;
        }

        public void RevokeFirstTeamStarPass()
        {
            Contract.Requires(FirstTeamLineup != null);
            Contract.Requires(FirstTeamScoring != null);
            RevokeTeamStarPass(FirstTeamLineup, FirstTeamScoring);
        }

        public void RevokeSecondTeamStarPass()
        {
            Contract.Requires(SecondTeamLineup != null);
            Contract.Requires(SecondTeamScoring != null);
            RevokeTeamStarPass(SecondTeamLineup, SecondTeamScoring);
        }

        private void RevokeTeamStarPass(Lineup lineup, Scoring scoring)
        {
            Contract.Requires(lineup != null);
            Contract.Requires(scoring != null);
            lineup.StarPassed = false;
            scoring.StarPassed = false;
            scoring.StarPassedDuringPass = null;
        }
    }
}
