﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel.DataAnnotations;
using System.ComponentModel.DataAnnotations.Schema;
using System.Diagnostics.Contracts;
using System.Linq;
using System.Text;

namespace Derbster.Domain
{
    [Serializable]
    public class Match : IEnumerable<Jam>
    {
        private const string Codes = "ABCEFGHILMNOPSXZ";

        [ContractVerification(false)]
        internal Match()
        {
            
        }

        public Match(bool isRegulation, bool isSanctioned, Roster firstTeamRoster, Roster secondTeamRoster)
            : this(isRegulation, isSanctioned, firstTeamRoster, secondTeamRoster, null, null, null)
        {
            Contract.Requires(firstTeamRoster != null);
            Contract.Requires(secondTeamRoster != null);
            Contract.Requires((isSanctioned && isRegulation) || (!isSanctioned));
            Contract.Ensures(Jams != null);
        }

        public Match(bool isRegulation, bool isSanctioned, Roster firstTeamRoster, Roster secondTeamRoster, string name)
            : this(isRegulation, isSanctioned, firstTeamRoster, secondTeamRoster, name, null, null)
        {
            Contract.Requires(firstTeamRoster != null);
            Contract.Requires(secondTeamRoster != null);
            Contract.Requires((isSanctioned && isRegulation) || (!isSanctioned));
            Contract.Ensures(Jams != null);
        }

        public Match(bool isRegulation, bool isSanctioned, Roster firstTeamRoster, Roster secondTeamRoster, string name, string where)
            : this(isRegulation, isSanctioned, firstTeamRoster, secondTeamRoster, name, where, null)
        {
            Contract.Requires(firstTeamRoster != null);
            Contract.Requires(secondTeamRoster != null);
            Contract.Requires((isSanctioned && isRegulation) || (!isSanctioned));
            Contract.Ensures(Jams != null);
        }

        public Match(bool isRegulation, bool isSanctioned, Roster firstTeamRoster, Roster secondTeamRoster, string name, string where, DateTime? when)
        {
            Contract.Requires(firstTeamRoster != null);
            Contract.Requires(secondTeamRoster != null);
            Contract.Requires((isSanctioned && isRegulation) || (!isSanctioned));
            Contract.Ensures(Jams != null);

            CurrentJamNumber = 0;
            CurrentPeriod = 0;
            IsRegulation = isRegulation;
            IsSanctioned = isSanctioned;
            FirstTeamRoster = firstTeamRoster;
            SecondTeamRoster = secondTeamRoster;
            Name = name;
            Where = where;
            When = when;

            Jams = new HashSet<Jam>();
            OfficialRoster = new OfficialRoster();
            Penalties = new HashSet<Penalty>();
            BoxEvents = new HashSet<BoxEvent>();  
        }

        [Key]
        public string MatchId { get; set; }

        public bool IsRegulation { get; set; }
        public bool IsSanctioned { get; set; }

        public Roster FirstTeamRoster { get; private set; }
        public Roster SecondTeamRoster { get; private set; }
        public OfficialRoster OfficialRoster { get; private set; }

        public HashSet<Jam> Jams { get; private set; }
        
        public string Name { get; set; }
        public DateTime? When { get; set; }
        public string Where { get; set; }

        public int CurrentJamNumber { get; private set; }
        public int CurrentPeriod { get; private set; }
        public bool HasStarted { get; private set; }
        public bool HasFinished { get; private set; }
        public bool JamInProgress { get; private set; }
        public bool PeriodInProgress { get; private set; }

        public HashSet<Penalty> Penalties { get; private set; }
        public HashSet<BoxEvent> BoxEvents { get; private set; }

        [NotMapped]
        public Jam CurrentJam
        {
            get
            {
                Contract.Requires(Jams != null);
                return Jams.FirstOrDefault(j => j.Period == CurrentPeriod && j.Number == CurrentJamNumber);
            }
        }

        public void StartMatch()
        {
            Contract.Requires(!JamInProgress);
            Contract.Requires(!HasStarted);
            Contract.Requires(!HasFinished);
            Contract.Requires(!PeriodInProgress);

            HasStarted = true;
        }
        
        public void UndoStartMatch()
        {
            Contract.Requires(!JamInProgress);
            Contract.Requires(HasStarted);
            Contract.Requires(!HasFinished);
            Contract.Requires(!PeriodInProgress);

            CurrentPeriod = 0;
            HasStarted = false;
        }

        public void EndMatch()
        {
            Contract.Requires(!JamInProgress);
            Contract.Requires(HasStarted);
            Contract.Requires(!HasFinished);
            Contract.Requires(!PeriodInProgress);

            CurrentPeriod = 0;
            HasStarted = false;
            HasFinished = true;
        }

        public void UndoEndMatch()
        {
            Contract.Requires(!JamInProgress);
            Contract.Requires(!HasStarted);
            Contract.Requires(HasFinished);
            Contract.Requires(!PeriodInProgress);

            HasFinished = false;
            HasStarted = true;
        }

        public void StartPeriod()
        {
            Contract.Requires(HasStarted);
            Contract.Requires(!HasFinished);
            Contract.Requires(!PeriodInProgress);
            Contract.Requires(CurrentPeriod == 0 || CurrentPeriod == 1);

            CurrentPeriod++;
            PeriodInProgress = true;
            
        }

        public void UndoStartPeriod()
        {
            Contract.Requires(HasStarted);
            Contract.Requires(!HasFinished);
            Contract.Requires(PeriodInProgress);
            Contract.Requires(CurrentPeriod <= 1);

            CurrentPeriod--;
            PeriodInProgress = false;
        }

        public void EndPeriod()
        {
            Contract.Requires(HasStarted);
            Contract.Requires(!HasFinished);
            Contract.Requires(PeriodInProgress);
            Contract.Requires(CurrentPeriod == 1 || CurrentPeriod == 2);

            CurrentJamNumber = 0;
            PeriodInProgress = false;
        }

        public void UndoEndPeriod()
        {
            Contract.Requires(HasStarted);
            Contract.Requires(!HasFinished);
            Contract.Requires(!PeriodInProgress);
            Contract.Requires(Jams != null);
            Contract.Requires(CurrentPeriod == 1 || CurrentPeriod == 2);

            CurrentJamNumber = Jams.Count(j => j.Period == CurrentPeriod);
            PeriodInProgress = true;
        }

        public Jam StartJam(Lineup firstLineup, Lineup secondLineup)
        {
            Contract.Requires(firstLineup != null);
            Contract.Requires(secondLineup != null);
            Contract.Requires(CurrentPeriod == 1 || CurrentPeriod == 2);
            Contract.Requires(CurrentJamNumber >= 0);
            Contract.Requires(HasStarted);
            Contract.Requires(!HasFinished);
            Contract.Requires(PeriodInProgress);
            Contract.Requires(!JamInProgress);
            Contract.Requires(Jams != null);

            CurrentJamNumber++;
            JamInProgress = true;
            var newJam = new Jam(this, CurrentPeriod, CurrentJamNumber,firstLineup,secondLineup);
            Jams.Add(newJam);
            return newJam;
        }

        public void UndoStartJam()
        {
            Contract.Requires(CurrentJamNumber >= 0);
            Contract.Requires(HasStarted);
            Contract.Requires(!HasFinished);
            Contract.Requires(PeriodInProgress);
            Contract.Requires(JamInProgress);
            Contract.Requires(Jams != null);

            Jams.RemoveWhere(j => j.Period == CurrentPeriod && j.Number == CurrentJamNumber);
            CurrentJamNumber--;
            JamInProgress = false;
        }

        public void EndJam()
        {
            Contract.Requires(HasStarted);
            Contract.Requires(!HasFinished);
            Contract.Requires(PeriodInProgress);
            Contract.Requires(JamInProgress);

            JamInProgress = false;
        }

        public void UndoEndJam()
        {
            Contract.Requires(HasStarted);
            Contract.Requires(!HasFinished);
            Contract.Requires(PeriodInProgress);
            Contract.Requires(!JamInProgress);

            JamInProgress = true;
        }

        public Penalty SkaterPenalty(Skater skater, string code)
        {
            Contract.Requires(skater != null);
            Contract.Requires(!String.IsNullOrWhiteSpace(code));
            Contract.Requires(Codes.Contains(code));
            Contract.Requires(CurrentJamNumber >= 1);
            Contract.Requires(CurrentPeriod == 1 || CurrentPeriod == 2);
            Contract.Requires(Penalties != null);

            var penalty = new Penalty(skater, code, CurrentPeriod, CurrentJamNumber);
            Penalties.Add(penalty);
            skater.State = SkaterState.InQueue;
            return penalty;
        }

        public string CorrectPenalty(Skater skater, int period, int jamNumber, string code)
        {
            Contract.Requires(skater != null);
            Contract.Requires(!String.IsNullOrWhiteSpace(code));
            Contract.Requires(Codes.Contains(code));
            Contract.Requires(period > 0 && period <= 2);
            Contract.Requires(jamNumber > 0);
            Contract.Requires(Penalties != null);

            var penalty =
                Penalties.SingleOrDefault(s => s.Skater == skater && s.Period == period && s.JamNumber == jamNumber);

            if (penalty == null)
                throw new IntegrityException("Skater penalty was not found.");

            string oldCode = penalty.Code;
            penalty.Code = code;
            return oldCode;
        }

        public string CorrectLastPenalty(Skater skater, string code)
        {
            Contract.Requires(skater != null);
            Contract.Requires(!String.IsNullOrWhiteSpace(code));
            Contract.Requires(Codes.Contains(code));
            Contract.Requires(Penalties != null);

            var skaterPenalities = Penalties.Where(s => s.Skater == skater).ToList();
            if (!skaterPenalities.Any())
                throw new IntegrityException("Skater has no penalities.");

            var lastPenalty = skaterPenalities.OrderBy(s => s.Period).ThenBy(s => s.JamNumber).Last();
            string oldCode = lastPenalty.Code;
            lastPenalty.Code = code;
            return oldCode;
        }

        public string RemoveLastPenalty(Skater skater)
        {
            Contract.Requires(skater != null);
            Contract.Requires(Penalties != null);

            var skaterPenalities = Penalties.Where(s => s.Skater == skater).ToList();
            if (!skaterPenalities.Any())
                throw new IntegrityException("Skater has no penalities.");

            var lastPenalty = skaterPenalities.OrderBy(s => s.Period).ThenBy(s => s.JamNumber).Last();
            string code = lastPenalty.Code;
            skater.State = SkaterState.Ready;
            Penalties.Remove(lastPenalty);
            return code;
        }

        public void AddPenalty(Penalty penalty)
        {
            Contract.Requires(penalty != null);
            Contract.Requires(penalty.Skater != null);
            Contract.Requires(Penalties != null);

            if (!Penalties.Contains(penalty))
            {
                Penalties.Add(penalty);
            }

            throw new IntegrityException("Penalty already exists.");
        }

        public void RemovePenalty(Penalty penalty)
        {
            Contract.Requires(penalty != null);
            Contract.Requires(penalty.Skater != null);
            Contract.Requires(Penalties != null);
            
            if (Penalties.Contains(penalty))
            {
                penalty.Skater.State = SkaterState.Ready;
                Penalties.Remove(penalty);
            }

            throw new IntegrityException("Invalid Penalty.");
        }

        public BoxEvent SkaterSeated(Skater skater)
        {
            Contract.Requires(skater != null);
            Contract.Requires(FirstTeamRoster != null);
            Contract.Requires(SecondTeamRoster != null);
            Contract.Requires(BoxEvents != null);
            Contract.Requires(CurrentJam != null);

            int pass;
             
            // Find opposing skating pass
            if (FirstTeamRoster.Members.Contains(skater))
            {
                pass = CurrentJam.SecondTeamScoring.Passes.Count+1;
            }
            else if (SecondTeamRoster.Members.Contains(skater))
            {
                pass = CurrentJam.FirstTeamScoring.Passes.Count+1;
            }
            else
            {
                throw new IntegrityException("Skater is not part of the lineup");
            }
            var boxEvent = new BoxEvent(skater, true, false, CurrentPeriod, CurrentJamNumber,pass);
            BoxEvents.Add(boxEvent);
            skater.State = SkaterState.InBox;
            return boxEvent;
        }

        public BoxEvent SkaterReleased(Skater skater)
        {
            Contract.Requires(skater != null);
            Contract.Requires(FirstTeamRoster != null);
            Contract.Requires(SecondTeamRoster != null);
            Contract.Requires(BoxEvents != null);
            Contract.Requires(CurrentJam != null);

            int pass;

            // Find opposing skating pass
            if (FirstTeamRoster.Members.Contains(skater))
            {
                pass = CurrentJam.SecondTeamScoring.Passes.Count + 1;
            }
            else if (SecondTeamRoster.Members.Contains(skater))
            {
                pass = CurrentJam.FirstTeamScoring.Passes.Count + 1;
            }
            else
            {
                throw new IntegrityException("Skater is not part of the lineup");
            }
            var boxEvent = new BoxEvent(skater, false, true, CurrentPeriod, CurrentJamNumber,pass);
            BoxEvents.Add(boxEvent);
            skater.State = SkaterState.Ready;
            return boxEvent;
        }

        public void UndoBoxEvent(BoxEvent boxEvent)
        {
            Contract.Requires(boxEvent != null);
            Contract.Requires(boxEvent.Skater != null);
            if (BoxEvents.Contains(boxEvent))
            {
                boxEvent.Skater.State = boxEvent.Released ? SkaterState.InBox : SkaterState.Ready;
                BoxEvents.Remove(boxEvent);
            }

            throw new IntegrityException("Invalid Box Event");
        }

        [Pure]
        public bool VerifyMatch()
        {
            if (IsSanctioned && !IsRegulation) 
                return false;

            if (!HasFinished)
                throw new IntegrityException("A match can only be verified if it is finished.");

            foreach (var jam in Jams)
            {
                if (jam.FirstTeamLineup == null || jam.FirstTeamScoring == null)
                    return false;
                if (jam.SecondTeamLineup == null || jam.SecondTeamScoring == null)
                    return false;
                if (jam.FirstTeamScoring.Jammer == null || jam.FirstTeamScoring.Jammer == null)
                    return false;
                if (jam.FirstTeamScoring.NoPass && jam.FirstTeamScoring.Passes.Count > 0)
                    return false;
                if (jam.SecondTeamScoring.NoPass && jam.SecondTeamScoring.Passes.Count > 0)
                    return false;
                if (jam.FirstTeamScoring.Passes.Any(sp => sp.Pass <= 0 || sp.Score < 0 || sp.Score > 6))
                    return false;
                if (jam.SecondTeamScoring.Passes.Any(sp => sp.Pass <= 0 || sp.Score < 0 || sp.Score > 6))
                    return false;
            }

            
            if (Penalties.Any(penalty => penalty.Code == null || !Codes.Contains(penalty.Code)))
                return false;

            foreach (var skater in Penalties.Select(p => p.Skater))
            {
                if ((Penalties.Count(p => skater == p.Skater)) !=
                    BoxEvents.Count(be => skater == be.Skater && be.Seated))
                    return false;
                if ((Penalties.Count(p => skater == p.Skater)) !=
                    BoxEvents.Count(be => skater == be.Skater && be.Released))
                    return false;
            }

            return true;
        }

        public IEnumerator<Jam> GetEnumerator()
        {
            Contract.Assume(Jams != null);
            return this.Jams.OrderBy(j => j.Period).ThenBy(j => j.Number).GetEnumerator();
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            Contract.Assume(Jams != null);
            return this.GetEnumerator();
        }
    }
}
